import dataclasses
from typing import List, Optional, Set

from botocore.client import BaseClient


@dataclasses.dataclass
class InstanceInfo:
    instance_id: str
    instance_type: str
    instance_az: str
    network_nodes: List[str]  # order: farthest from instance -> closest to instance


def iter_instance_info(
    ec2_client: BaseClient,
    accepted_node_instance_types: Set[str],
    accepted_node_availability_zones: Set[str],
    accepted_instance_ids: Optional[Set[str]] = None,
):
    # ref: https://docs.aws.amazon.com/cli/latest/reference/ec2/describe-instance-topology.html
    paginator = ec2_client.get_paginator("describe_instance_topology")

    # we have to filter in this roundabout way instead of giving a set of instance IDs
    # because API doesn't accept >100 instance IDs
    instance_filters = [
        {
            "Name": "instance-type",
            "Values": list(accepted_node_instance_types),
        },
        {
            "Name": "availability-zone",
            "Values": list(accepted_node_availability_zones),
        },
    ]

    for page in paginator.paginate(Filters=instance_filters, PaginationConfig={"PageSize": 100}):
        for instance_info in page["Instances"]:
            if (
                accepted_instance_ids is None
                or instance_info["InstanceId"] in accepted_instance_ids
            ):
                yield InstanceInfo(
                    instance_id=instance_info["InstanceId"],
                    instance_type=instance_info["InstanceType"],
                    instance_az=instance_info["AvailabilityZone"],
                    network_nodes=instance_info["NetworkNodes"],
                )



import dataclasses
from enum import Enum

from .configurations import ActivationCheckpointingType
from .data import Size, TensorRepr
from .math import ceil_divide, safe_divide


@dataclasses.dataclass
class ParallelConfig:
    class ZeroLevel(int, Enum):
        NONE = 0
        PARTITION_OPTIMIZER = 1
        PARTITION_GRADIENTS = 2
        PARTITION_PARAMETERS = 3

    tp: int  # tensor parallel degree
    pp: int  # pipeline parallel degree
    dp: int  # data parallel degree

    vpp: int  # virtual pipeline parallel degree

    sp_enabled: bool  # sequence parallel

    zero_level: ZeroLevel

    def mp_degree(self) -> int:
        return self.tp * self.pp

    def world_size(self) -> int:
        return self.tp * self.pp * self.dp


class DistributedAdamOptimizerStates:
    """Optimizer states from Apex DistributedFusedAdam.
    see: https://github.com/NVIDIA/Megatron-LM/blob/main/docs/source/distrib_optimizer.md
    and: https://github.com/NVIDIA/apex/blob/master/apex/contrib/optimizers/distributed_fused_adam.py

    the distributed optimizer recieves a set of parameters to manage that are already
    model parallel partitioned. Internally, it will additionally partition states
    over DP, so optimizer states end up being partitioned over MP * DP, but
    the diestributed optimizer doesn't have any concept of model parallelism.

    NOTE: shards will be larger in reality due to alignment requirements and
    unfilled buckets.
    """

    def __init__(self, n_mp_params: int, store_param_remainders: bool, dp: int) -> None:
        self.param_shard = TensorRepr(
            unpartitioned_shape=(n_mp_params,),
            partition_dim=0,
            partition_degree=dp,
            # apex has a optimization to avoid storing information that's redundant
            # between fp32 and fp16 weights. it can instead store an extra 16
            # bits of precision, which can be combined with bf16 weights to yield
            # fp32 weights
            bits_per_elt=16 if store_param_remainders else 32,
            enforce_evenly_partitionable=False,
        )
        self.exp_avg_shard = TensorRepr(
            unpartitioned_shape=(n_mp_params,),
            partition_dim=0,
            partition_degree=dp,
            bits_per_elt=32,
            enforce_evenly_partitionable=False,
        )
        self.exp_avg_sq_shard = TensorRepr(
            unpartitioned_shape=(n_mp_params,),
            partition_dim=0,
            partition_degree=dp,
            bits_per_elt=32,
            enforce_evenly_partitionable=False,
        )
        # reused for various purposes but at one point it needs to hold
        # all (i.e. not DP partitioned) parameter gradients, which are accumulated
        # to until they are reduce-scattered to a gradient shard during the final
        # microbatch.
        # basically ZeRO2, but with no reduce-scatter between gradient accumulations.
        self.grad_buffer = TensorRepr(
            unpartitioned_shape=(n_mp_params,),
            partition_dim=None,
            partition_degree=1,
            bits_per_elt=32,
        )

    def total_bytes(self, partitioned: bool) -> int:
        return _sum(
            self.param_shard.size(partitioned=partitioned).bytes(),
            self.exp_avg_shard.size(partitioned=partitioned).bytes(),
            self.exp_avg_sq_shard.size(partitioned=partitioned).bytes(),
            self.grad_buffer.size(partitioned=partitioned).bytes(),
        )


@dataclasses.dataclass
class ModelStates:
    """Tracks persistent tensor allocations kept throughput training."""

    params_shard: TensorRepr
    opt_states: DistributedAdamOptimizerStates

    def total_bytes(self, partitioned: bool) -> int:
        return self.params_shard.size(
            partitioned=partitioned
        ).bytes() + self.opt_states.total_bytes(partitioned=partitioned)

    def __repr__(self) -> str:
        params_bytes = self.params_shard.size(partitioned=True).bytes()
        opt_states_bytes = self.opt_states.total_bytes(partitioned=True)

        return "\n".join(
            [
                f"params          : {self.params_shard.size(partitioned=True)}",
                f"params (opt)    : {self.opt_states.param_shard.size(partitioned=True)}",
                f"exp_avg         : {self.opt_states.exp_avg_shard.size(partitioned=True)}",
                f"exp_avg_squared : {self.opt_states.exp_avg_sq_shard.size(partitioned=True)}",
                f"grad_buffer     : {self.opt_states.grad_buffer.size(partitioned=True)}",
                f"TOTAL           : {(params_bytes + opt_states_bytes) / (1024 ** 3):.2f}GiB",
            ]
        )


def _sum(*summands: int) -> int:
    return sum(list(summands))


@dataclasses.dataclass
class ThreeDParallelModel:
    """Representation of a 3D parallel transformer model."""

    parallelism_cfg: ParallelConfig

    sequence_len: int
    microbatch_sz: int

    hidden_sz: int

    n_layers: int

    n_q_heads: int
    n_kv_heads: int  # num_query_groups if GQA
    head_dim: int

    inter_sz: int
    glu: bool

    rotary_embed: bool

    dropout: bool

    vocab_sz: int
    tie_embeddings: bool

    act_ckpting_type: ActivationCheckpointingType

    bucket_size_bytes: int

    # TODO. assuming mixed precision here.
    bits_per_parameter: int = 16
    bits_per_grad: int = 32
    bits_per_optim_state: int = 32

    def __post_init__(self) -> None:
        if self.n_layers % (self.parallelism_cfg.pp * self.parallelism_cfg.vpp) != 0:
            raise ValueError(
                f"number of layers {self.n_layers} is not divisible by the product "
                f"of PP={self.parallelism_cfg.pp} and VPP={self.parallelism_cfg.vpp}"
            )

        self.embed_weight = TensorRepr(
            unpartitioned_shape=(self.hidden_sz, self.vocab_sz),
            partition_dim=1,
            partition_degree=self.parallelism_cfg.tp,
            bits_per_elt=self.bits_per_parameter,
        )
        self.norm1_weight = TensorRepr(
            unpartitioned_shape=(self.hidden_sz,),
            partition_dim=None,
            partition_degree=1,
            bits_per_elt=self.bits_per_parameter,
        )
        self.qkv_weight = TensorRepr(
            unpartitioned_shape=(
                self.hidden_sz,
                (self.n_q_heads + 2 * self.n_kv_heads) * self.head_dim,
            ),
            partition_dim=1,
            partition_degree=self.parallelism_cfg.tp,
            bits_per_elt=self.bits_per_parameter,
        )
        self.attn_out_weight = TensorRepr(
            unpartitioned_shape=(self.hidden_sz, self.hidden_sz),
            partition_dim=0,
            partition_degree=self.parallelism_cfg.tp,
            bits_per_elt=self.bits_per_parameter,
        )
        self.mlp_up_weight = TensorRepr(
            # following common practice of merging up + gate matmuls in the event
            # we're using GLU.
            unpartitioned_shape=(
                self.hidden_sz,
                (self.inter_sz * 2) if self.glu else self.inter_sz,
            ),
            partition_dim=1,
            partition_degree=self.parallelism_cfg.tp,
            bits_per_elt=self.bits_per_parameter,
        )
        self.mlp_down_weight = TensorRepr(
            unpartitioned_shape=(self.inter_sz, self.hidden_sz),
            partition_dim=0,
            partition_degree=self.parallelism_cfg.tp,
            bits_per_elt=self.bits_per_parameter,
        )
        self.norm2_weight = TensorRepr(
            unpartitioned_shape=(self.hidden_sz,),
            partition_dim=None,
            partition_degree=1,
            bits_per_elt=self.bits_per_parameter,
        )

        n_params_most_loaded_pp_stage = self.__get_embedding_or_lm_head_size(
            partitioned=True
        ) + self.layers_per_pp_stage() * self.__get_transformer_block_n_params(partitioned=True)

        if self.parallelism_cfg.zero_level != ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER:
            raise NotImplementedError

        self.states = ModelStates(
            params_shard=TensorRepr(
                unpartitioned_shape=(self.get_total_n_params(partitioned=False),),
                partition_dim=0,
                partition_degree=self.parallelism_cfg.mp_degree(),
                bits_per_elt=self.bits_per_parameter,
                enforce_evenly_partitionable=False,
            ),
            opt_states=DistributedAdamOptimizerStates(
                n_mp_params=n_params_most_loaded_pp_stage,
                store_param_remainders=True,  # TODO. should be configurable
                dp=self.parallelism_cfg.dp,
            ),
        )

    def get_single_microbatch_fwd_flops(self) -> float:
        return (
            2  # FLOPs/MAC
            * 1  # factor for forward only (1 GEMMs per op)
            * self.microbatch_sz
            * self.sequence_len
            * self.get_total_n_params(partitioned=False)
        )

    def get_single_microbatch_bwd_flops(self) -> float:
        return (
            2  # FLOPs/MAC
            * 2  # factor for backward only (2 GEMMs per op)
            * self.microbatch_sz
            * self.sequence_len
            * self.get_total_n_params(partitioned=False)
        )

    def get_total_n_params(self, partitioned: bool) -> int:
        embedding_or_lm_head_n_params = self.__get_embedding_or_lm_head_size(
            partitioned=partitioned
        )
        transformer_block_n_params = self.__get_transformer_block_n_params(partitioned=partitioned)

        if partitioned:
            # we'll give the number of parameters on the most heavily loaded pipeline stage
            return (
                embedding_or_lm_head_n_params
                + safe_divide(self.n_layers, self.parallelism_cfg.pp) * transformer_block_n_params
            )
        else:
            embedding_or_lm_head_factor = 1 if self.tie_embeddings else 2
            return (
                embedding_or_lm_head_factor * embedding_or_lm_head_n_params
                + self.n_layers * transformer_block_n_params
            )

    def __get_embedding_or_lm_head_size(self, partitioned: bool) -> int:
        return self.embed_weight.size(partitioned=partitioned).numel()

    def __get_transformer_block_n_params(self, partitioned: bool) -> int:
        return _sum(
            self.norm1_weight.size(partitioned=partitioned).numel(),
            self.qkv_weight.size(partitioned=partitioned).numel(),
            self.attn_out_weight.size(partitioned=partitioned).numel(),
            self.norm2_weight.size(partitioned=partitioned).numel(),
            self.mlp_up_weight.size(partitioned=partitioned).numel(),
            self.mlp_down_weight.size(partitioned=partitioned).numel(),
        )

    def activation_size_per_microbatch_per_layer(self) -> Size:
        return Size(
            numel=self.__activation_numel_per_microbatch_per_layer(),
            bits_per_element=self.bits_per_parameter,
        )

    def vpp_penalty(self) -> float:
        """interleaved schedule requires storing activations for (1 + (p - 1)/pm)
        more layers."""
        if self.parallelism_cfg.vpp == 1:
            return 1.0

        return 1 + (self.parallelism_cfg.pp - 1) / (
            self.parallelism_cfg.pp * self.parallelism_cfg.vpp
        )

    def layers_per_pp_stage(self) -> int:
        return safe_divide(self.n_layers, self.parallelism_cfg.pp)

    def grad_bucket_numel(self) -> int:
        return ceil_divide(self.bucket_size_bytes, self.bits_per_grad // 8)

    def __activation_numel_per_microbatch_per_layer(self) -> int:
        """
        See: Reducing Activation Recomputation in Large Transformer Models
        https://arxiv.org/pdf/2205.05198.pdf
        """
        sbh = self.sequence_len * self.microbatch_sz * self.hidden_sz
        sbi = self.sequence_len * self.microbatch_sz * self.inter_sz
        sbq = self.sequence_len * self.microbatch_sz * self.n_q_heads * self.head_dim
        sbk = self.sequence_len * self.microbatch_sz * self.n_kv_heads * self.head_dim
        sbv = self.sequence_len * self.microbatch_sz * self.n_kv_heads * self.head_dim

        if self.act_ckpting_type == ActivationCheckpointingType.FULL:
            return self.__sp_partition_if_on(sbh)  # just the block input
        elif self.act_ckpting_type == ActivationCheckpointingType.SUPER_SELECTIVE:
            return _sum(
                # LAYERNORM 1
                # - output is recomputed
                # QKV (col parallel linear)
                self.__tp_partition(sbq),  # Q - attn input
                self.__tp_partition(sbk),  # K - attn input
                self.__tp_partition(sbv),  # V - attn input
                # ROTARY EMBEDDINGS
                self.__tp_partition(sbq) if self.rotary_embed else 0,  # Q rotary
                self.__tp_partition(sbk) if self.rotary_embed else 0,  # K rotary
                # SELF ATTENTION
                # - skipping intermediates (checkpointed by FlashAttention)
                self.__tp_partition(sbh),  # attn output
                # DOWN PROJ
                # - output deallocated (dropout doesn't need to store)
                # DROPOUT
                # - dropout mask recomputed
                # - deallocated - residual doesn't need to store
                # RESIDUAL
                self.__sp_partition_if_on(sbh),  # stored by norm2, residual2
                # LAYERNORM 2
                # output is recomputed
                # MLP UP/GATE (col parallel linear)
                self.__tp_partition(2 * sbi if self.glu else sbi),  # SwiGLU input
                # SwiGLU
                # - output is recomputed
                # MLP DOWN (row parallel linear)
                # - output deallocated - dropout doesn't need to store
                # DROPOUT
                # - dropout mask recomputed
                # - output deallocated - residual doesn't need to store
                # RESIDUAL
                self.__sp_partition_if_on(sbh),  # needed by next norm1, residual1
            )
        elif self.act_ckpting_type == ActivationCheckpointingType.SELECTIVE:
            return _sum(
                # LAYERNORM 1
                self.__sp_partition_if_on(sbh),  # output - QKV input
                # QKV PROJ (col parallel linear)
                self.__tp_partition(sbq),  # Q - attn input
                self.__tp_partition(sbk),  # K - attn input
                self.__tp_partition(sbv),  # V - attn input
                # ROTARY EMBEDDINGS
                self.__tp_partition(sbq) if self.rotary_embed else 0,  # Q rotary
                self.__tp_partition(sbk) if self.rotary_embed else 0,  # K rotary
                # SELF ATTENTION
                # - skipping intermediates (checkpointed by FlashAttention)
                self.__tp_partition(sbh),  # needed by down proj
                # DOWN PROJ
                # - output deallocated (dropout doesn't need to store)
                # DROPOUT
                self.__sp_partition_if_on(int(0.5 * sbh)) if self.dropout else 0,  # dropout mask
                # -  output deallocated: residual doesn't need to store
                # RESIDUAL
                self.__sp_partition_if_on(sbh),  # needed by norm2, resid2
                # LAYERNORM 2
                self.__sp_partition_if_on(sbh),  # up/gate input
                # MLP UP/GATE (col parallel linear)
                self.__tp_partition(2 * sbi if self.glu else sbi),  # SwiGLU input
                # SwiGLU
                self.__tp_partition(sbi),  # SiLU output
                self.__tp_partition(sbi),  # gate output
                # MLP DOWN (row parallel linear)
                # - output deallocated - dropout doesn't need to store
                # DROPOUT
                self.__sp_partition_if_on(int(0.5 * sbh)) if self.dropout else 0,  # dropout mask
                # - output deallocated - residual doesn't need to store
                # RESIDUAL
                # input to next norm1, resid1
                self.__sp_partition_if_on(sbh),
            )
        elif self.act_ckpting_type == ActivationCheckpointingType.NONE:
            raise NotImplementedError(
                "not yet implemented. selective will give a good approximation "
                "if you use flash attention"
            )
        else:
            raise ValueError(f"unhandled checkpointing_type={self.act_ckpting_type}")

    def __tp_partition(self, unpartitoned_numel: int) -> int:
        return safe_divide(unpartitoned_numel, self.parallelism_cfg.tp)

    def __sp_partition_if_on(self, unpartitioned_numel: int) -> int:
        if not self.parallelism_cfg.sp_enabled:
            return unpartitioned_numel
        return safe_divide(unpartitioned_numel, self.parallelism_cfg.tp)



"""
given a kubernetes pod name prefix for some compute cluster, retrieve AWS network
topology info and plot.
"""

import re
from argparse import ArgumentParser

import boto3
from pyvis.network import Network

import dlcalc.utils.cluster.ec2
import dlcalc.utils.cluster.kubernetes

pod_id_pattern = re.compile(r"worker-(\d+)")


def _dict_to_label(dict: dict) -> str:
    kv_pairs = []
    for key, value in dict.items():
        kv_pairs.append(f"{key}: {value}")
    return "\n".join(kv_pairs)


def main() -> None:
    parser = ArgumentParser(__doc__)
    parser.add_argument(
        "search_prefix",
        help="kubernetes pod name prefix to use when filtering pods returned by `kubectl get pods`",
    )
    args = parser.parse_args()

    job_search_prefix = args.search_prefix

    ######################################################################################
    # STEP 1: Get the list of instance IDs involved in a kubernetes job
    ######################################################################################
    cluster_members = dlcalc.utils.cluster.kubernetes.get_kubernetes_cluster_members(
        job_search_prefix
    )

    ######################################################################################
    # STEP 2: Get topology of the instances acquired earlier.
    ######################################################################################
    instance_ids = [m.node_instance_id for m in cluster_members]
    (cluster_region,) = set(m.node_region for m in cluster_members)
    cluster_instance_types = list(set(m.node_instance_type for m in cluster_members))
    cluster_azs = list(set(m.node_az for m in cluster_members))
    instance_id_to_job_member = {m.node_instance_id: m for m in cluster_members}

    print(f"getting topology info for instance IDs {instance_ids}")
    ec2_client = boto3.client("ec2", region_name=cluster_region)

    net = Network(directed=True)
    for instance_info in dlcalc.utils.cluster.ec2.iter_instance_info(
        ec2_client=ec2_client,
        accepted_node_instance_types=cluster_instance_types,
        accepted_node_availability_zones=cluster_azs,
        accepted_instance_ids=instance_ids,
    ):
        job_member = instance_id_to_job_member[instance_info.instance_id]

        net.add_node(
            instance_info.instance_id,
            title=_dict_to_label(
                {
                    "PodName": job_member.pod_name,
                    "WorkerId": job_member.worker_id,
                    "InstanceType": instance_info.instance_type,
                    "AvailabilityZone": instance_info.instance_az,
                }
            ),
            color="blue",
            physics=False,
        )

        for network_node in instance_info.network_nodes:
            net.add_node(network_node, color="red", physics=False)

        # order of network nodes: farthest from instance -> closest to instance
        chain = [*instance_info.network_nodes, instance_info.instance_id]
        for edge_src_idx in range(len(chain) - 1):
            edge_dst_idx = edge_src_idx + 1
            net.add_edge(chain[edge_src_idx], chain[edge_dst_idx], physics=False)

    net.show_buttons()
    net.show("graph.html", notebook=False)


if __name__ == "__main__":
    main()



"""simple utility to convert samples/second to tokens/day"""

from argparse import ArgumentParser


def main() -> None:
    parser = ArgumentParser(__doc__)
    parser.add_argument(
        "-t",
        "--samples-per-sec",
        type=float,
        required=True,
        help="training throughput in samples/sec",
    )
    parser.add_argument(
        "-s",
        "--seqlen",
        type=int,
        required=True,
        help="sequence length in tokens (i.e. tokens/sample)",
    )
    parser.add_argument(
        "-g",
        "--goodput",
        type=float,
        default=1.0,
        required=False,
        help="training goodput. expected to be in range [0.0, 1.0]",
    )
    args = parser.parse_args()

    samples_per_sec = args.samples_per_sec
    tokens_per_sample = args.seqlen
    goodput = args.goodput

    if not 0.0 <= goodput <= 1.0:
        raise ValueError(f"goodput expected to be in range [0.0, 1.0] but got {goodput}")

    tokens_per_day = samples_per_sec * tokens_per_sample * 60 * 60 * 24 * goodput

    print(
        f"{samples_per_sec} samples/sec "
        f"with sequence length {tokens_per_sample} "
        f"and {goodput * 100:.2f}% goodput "
        f"translates to {tokens_per_day * 1e-9:.2f}B tokens/day"
    )


if __name__ == "__main__":
    main()



"""
Reduce-Scatter:
* Ring: https://github.com/NVIDIA/nccl/blob/ab2b89c4c339bd7f816fbc114a4b05d386b66290/src/device/reduce_scatter.h#L12-L54

All-Gather:
* Ring: https://github.com/NVIDIA/nccl/blob/ab2b89c4c339bd7f816fbc114a4b05d386b66290/src/device/all_gather.h#L12-L64

All-Reduce (not relevant for comm patterns we have represented for now):
some info on tree AR: https://github.com/NVIDIA/nccl/issues/545

TODO. try to get estimates closer by accounting for NCCL protocols https://github.com/NVIDIA/nccl/issues/281
"""

from .data import Size
from .hardware import MachineSpec


def get_tp_reduce_scatter_comm_time_s(
    size: Size, n_participants: int, machine_spec: MachineSpec
) -> float:
    """assumes ring algorithm."""
    return _get_ring_tp_ag_or_rs_comm_time_s(
        size, n_participants=n_participants, machine_spec=machine_spec
    )


def get_tp_all_gather_comm_time_s(
    size: Size, n_participants: int, machine_spec: MachineSpec
) -> float:
    """assumes ring algorithm."""
    return _get_ring_tp_ag_or_rs_comm_time_s(
        size, n_participants=n_participants, machine_spec=machine_spec
    )


def get_dp_reduce_scatter_latency_term_s(
    n_participants: int,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    return _ring_ag_or_rs_latency_term_s(
        n_participants=n_participants,
        link_latency_s=machine_spec.inter_node_connect.latency_sec,
    )


def get_dp_reduce_scatter_bw_term_s(
    size: Size,
    n_participants: int,
    mp_degree_in_node: int,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    return _ring_ag_or_rs_bw_term_s(
        size,
        n_participants=n_participants,
        # full BW should be divided along all MP ranks within a single node, since
        # they are each participating in their own DP collectives.
        unidirectional_link_bw_bytes_per_sec=machine_spec.inter_node_connect.unidirectional_bw_bytes_per_sec
        // mp_degree_in_node,
    )


def get_dp_reduce_scatter_comm_time_s(
    size: Size,
    n_participants: int,
    mp_degree_in_node: int,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    latency_term = get_dp_reduce_scatter_latency_term_s(
        n_participants=n_participants,
        machine_spec=machine_spec,
    )

    bw_term = get_dp_reduce_scatter_bw_term_s(
        size,
        n_participants=n_participants,
        mp_degree_in_node=mp_degree_in_node,
        machine_spec=machine_spec,
    )

    return latency_term + bw_term


def get_dp_all_gather_latency_term_s(
    n_participants: int,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    return _ring_ag_or_rs_latency_term_s(
        n_participants=n_participants,
        link_latency_s=machine_spec.inter_node_connect.latency_sec,
    )


def get_dp_all_gather_bw_term_s(
    size: Size,
    n_participants: int,
    mp_degree_in_node: int,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    return _ring_ag_or_rs_bw_term_s(
        size,
        n_participants=n_participants,
        # full BW should be divided along all MP ranks within a single node, since
        # they are each participating in their own DP collectives.
        unidirectional_link_bw_bytes_per_sec=machine_spec.inter_node_connect.unidirectional_bw_bytes_per_sec
        // mp_degree_in_node,
    )


def get_dp_all_gather_comm_time_s(
    size: Size,
    n_participants: int,
    mp_degree_in_node: int,
    machine_spec: MachineSpec,
) -> float:
    """assumes ring algorithm."""
    latency_term = get_dp_reduce_scatter_latency_term_s(
        n_participants=n_participants,
        machine_spec=machine_spec,
    )

    bw_term = get_dp_reduce_scatter_bw_term_s(
        size,
        n_participants=n_participants,
        mp_degree_in_node=mp_degree_in_node,
        machine_spec=machine_spec,
    )

    return latency_term + bw_term


def _ring_ag_or_rs_latency_term_s(
    n_participants: int,
    link_latency_s: float,
) -> float:
    return (n_participants - 1) * link_latency_s


def _ring_ag_or_rs_bw_term_s(
    size: Size,
    n_participants: int,
    # NOTE: assumes duplex bw = 2x unidirectional
    unidirectional_link_bw_bytes_per_sec: int,
) -> float:
    phase_sent_data_size_bytes = int(size.bytes() / n_participants)
    return (n_participants - 1) * (
        phase_sent_data_size_bytes / unidirectional_link_bw_bytes_per_sec
    )


def _get_ring_tp_ag_or_rs_comm_time_s(
    size: Size,
    n_participants: int,
    machine_spec: MachineSpec,
) -> float:
    lat_term_s = _ring_ag_or_rs_latency_term_s(
        n_participants=n_participants,
        link_latency_s=machine_spec.intra_node_connect.latency_sec,
    )
    bw_term_s = _ring_ag_or_rs_bw_term_s(
        size,
        n_participants=n_participants,
        unidirectional_link_bw_bytes_per_sec=machine_spec.intra_node_connect.unidirectional_bw_bytes_per_sec,
    )

    return bw_term_s + lat_term_s



"""
evaluates how optimal a given training job's physical topology is (in terms of
network hops in each DP ring).
"""

from argparse import ArgumentParser

import boto3

import dlcalc.utils.cluster.ec2
import dlcalc.utils.cluster.kubernetes
import dlcalc.utils.cluster.topology
from dlcalc.utils.math import safe_divide


def main() -> None:
    parser = ArgumentParser(__doc__)
    parser.add_argument(
        "job_search_prefix",
        help="kubernetes pod name prefix to use when filtering pods returned by `kubectl get pods`",
    )
    parser.add_argument(
        "-t",
        "--tp-degree",
        type=int,
        required=True,
        help="tensor parallel (TP) degree",
    )
    parser.add_argument(
        "-p",
        "--pp-degree",
        type=int,
        required=True,
        help="pipeline parallel (PP) degree",
    )
    parser.add_argument(
        "-d",
        "--dp-degree",
        type=int,
        required=True,
        help="data parallel (DP) degree",
    )
    parser.add_argument(
        "-n",
        "--accelerators-per-node",
        type=int,
        default=8,
        help="number of accelerators per node",
    )
    args = parser.parse_args()
    job_search_prefix = args.job_search_prefix
    tp_degree = args.tp_degree
    pp_degree = args.pp_degree
    dp_degree = args.dp_degree
    accelerators_per_node = args.accelerators_per_node

    world_size = tp_degree * pp_degree * dp_degree
    expected_n_nodes = safe_divide(world_size, accelerators_per_node)

    assert tp_degree <= accelerators_per_node

    # we only care to optimize inter-node DP communication
    # (intra-node DP communication can't be optimized with better instance placement),
    # so we'll ignore the parts of the ring that traverse intra-node links.
    dp_ranks_per_node = safe_divide(accelerators_per_node, tp_degree)
    nodes_per_dp_ring = safe_divide(dp_degree, dp_ranks_per_node)
    print(
        f"DP={dp_degree} and {dp_ranks_per_node} DP ranks/node, "
        f"therefore there are {nodes_per_dp_ring} nodes per DP ring"
    )

    # everything we look at here is on the instance level.

    # if TP < accelerators per node, then there will be multiple DP ranks on a single
    # node. we'll ignore this intra-node DP communication as we just care about
    # improving inter-node DP communication
    # we're making the assumption that all TP groups are within a single node.
    # since we're looking at things at the instance level, there's only one
    # ring per PP rank.

    ######################################################################################
    # STEP 1: Get instance info involved in a kubernetes job
    ######################################################################################
    cluster_members = dlcalc.utils.cluster.kubernetes.get_kubernetes_cluster_members(
        job_search_prefix
    )
    if len(cluster_members) != expected_n_nodes:
        raise ValueError(
            f"with world size {world_size} "
            f"expected to find {expected_n_nodes} nodes "
            f"but found {len(cluster_members)}"
        )
    worker_id_to_instance_id = {m.worker_id: m.node_instance_id for m in cluster_members}
    (cluster_region,) = {m.node_region for m in cluster_members}
    cluster_instance_types = {m.node_instance_type for m in cluster_members}
    cluster_azs = {m.node_az for m in cluster_members}

    print(
        f"found {len(cluster_members)} matching pods, with "
        f"region: {cluster_region}, instance_types: {cluster_instance_types}, azs: {cluster_azs}"
    )

    node_id_to_node = dlcalc.utils.cluster.topology.build_instance_tree(
        ec2_client=boto3.client("ec2", region_name=cluster_region),
        accepted_node_instance_types=cluster_instance_types,
        accepted_node_availability_zones=cluster_azs,
        accepted_instance_ids=set(worker_id_to_instance_id.values()),
    )

    ######################################################################################
    # STEP 2: Determine connectivity path of DP rings
    ######################################################################################

    # find placement on physical topology
    for pp_rank in range(pp_degree):
        dp_ring_for_pp_rank = [None] * nodes_per_dp_ring
        for dp_rank in range(nodes_per_dp_ring):
            worker_id = pp_rank * nodes_per_dp_ring + dp_rank
            instance_id = worker_id_to_instance_id[worker_id]
            dp_ring_for_pp_rank[dp_rank] = node_id_to_node[instance_id]

        traversal_dist = dlcalc.utils.cluster.topology.traversal_distance_of_ring(
            dp_ring_for_pp_rank
        )
        print(
            f"DP ring for PP={pp_rank}: traversal distance: {traversal_dist} "
            f"-> avg traversal distance: {traversal_dist / nodes_per_dp_ring:.2f}"
        )
        print()


if __name__ == "__main__":
    main()



import json
from dataclasses import dataclass

EFA_LATENCY_S = 30e-6


@dataclass
class DeviceSpec:
    # TODO. make it depend on dtype
    peak_flops: int
    mem_bandwidth_bytes_per_sec: int
    mem_capacity_bytes: int


@dataclass
class LinkSpec:
    unidirectional_bw_bytes_per_sec: int
    latency_sec: float

    def __repr__(self) -> str:
        return json.dumps(
            {
                "unidirectional bw": f"{self.unidirectional_bw_bytes_per_sec * 1e-9:.2f} GBps",
                "latency": f"{self.latency_sec * 1e6} us",
            }
        )


@dataclass
class MachineSpec:
    n_devices: int
    device_spec: DeviceSpec
    intra_node_connect: LinkSpec
    inter_node_connect: LinkSpec

    @staticmethod
    def from_str(str: str) -> "MachineSpec":
        return {
            # A100 datasheet: https://www.nvidia.com/content/dam/en-zz/Solutions/Data-Center/a100/pdf/nvidia-a100-datasheet-us-nvidia-1758950-r4-web.pdf
            # https://aws.amazon.com/ec2/instance-types/p4/
            "p4d.24xlarge": MachineSpec(
                n_devices=8,
                # A100-40GiB SXM
                device_spec=DeviceSpec(
                    peak_flops=int(312e12),
                    # 40GiB HBM2
                    mem_bandwidth_bytes_per_sec=int(1555e9),
                    mem_capacity_bytes=40 * (1024**3),
                ),
                # NVLink 3
                intra_node_connect=LinkSpec(
                    unidirectional_bw_bytes_per_sec=int(300e9),
                    latency_sec=3e-6,
                ),
                # EFA v1 - 4 x 100 Gbps
                inter_node_connect=LinkSpec(
                    unidirectional_bw_bytes_per_sec=int(50e9),
                    latency_sec=EFA_LATENCY_S,
                ),
            ),
            # https://aws.amazon.com/ec2/instance-types/p4/
            "p4de.24xlarge": MachineSpec(
                n_devices=8,
                # A100-80GiB SXM
                device_spec=DeviceSpec(
                    peak_flops=int(312e12),
                    # 80GiB HBM2e
                    mem_bandwidth_bytes_per_sec=int(2039e9),
                    mem_capacity_bytes=40 * (1024**3),
                ),
                # NVLink 3
                intra_node_connect=LinkSpec(
                    unidirectional_bw_bytes_per_sec=int(300e9),
                    latency_sec=3e-6,
                ),
                # EFA v1 - 4 x 100 Gbps
                inter_node_connect=LinkSpec(
                    unidirectional_bw_bytes_per_sec=int(50e9),
                    latency_sec=EFA_LATENCY_S,
                ),
            ),
            # H100 datasheet: https://resources.nvidia.com/en-us-tensor-core/nvidia-tensor-core-gpu-datasheet
            # https://aws.amazon.com/ec2/instance-types/p5/
            "p5.48xlarge": MachineSpec(
                n_devices=8,
                # H100 SXM
                device_spec=DeviceSpec(
                    peak_flops=int(989e12),  # (ignore sparsity numbers)
                    # 80 GiB HBM3
                    mem_bandwidth_bytes_per_sec=int(3350e9),
                    mem_capacity_bytes=80 * (1024**3),
                ),
                # NVLink 4
                intra_node_connect=LinkSpec(
                    unidirectional_bw_bytes_per_sec=int(450e9),
                    latency_sec=3e-6,
                ),
                # EFA v2 - 32 x 100 Gbps
                inter_node_connect=LinkSpec(
                    unidirectional_bw_bytes_per_sec=int(400e9),
                    latency_sec=EFA_LATENCY_S,
                ),
            ),
            # https://aws.amazon.com/ec2/instance-types/trn1/
            # https://awsdocs-neuron.readthedocs-hosted.com/en/latest/general/arch/neuron-hardware/trn1-arch.html
            "trn1n.32xlarge": MachineSpec(
                n_devices=32,  # technically 16, but treating neuron cores as devices.
                # NeuronCore v2
                device_spec=DeviceSpec(
                    peak_flops=int(95e12),
                    # 16 GiB HBM2e
                    mem_bandwidth_bytes_per_sec=int(410e9),
                    mem_capacity_bytes=16 * (1024**3),
                ),
                # NeuronLink v2
                intra_node_connect=LinkSpec(
                    unidirectional_bw_bytes_per_sec=int(384e9),
                    latency_sec=float("inf"),  # TODO. not sure, figure out empirically
                ),
                # EFA v2
                inter_node_connect=LinkSpec(
                    unidirectional_bw_bytes_per_sec=int(200e9),
                    latency_sec=EFA_LATENCY_S,
                ),
            ),
        }[str]

    def total_flops(self) -> int:
        return self.device_spec.peak_flops * self.n_devices



import dataclasses
import json
import re
from typing import List, Set

import kubernetes
from kubernetes.client import CoreV1Api

pod_id_pattern = re.compile(r"worker-(\d+)")


@dataclasses.dataclass
class KubernetesJobMember:
    pod_name: str
    worker_id: int

    node_instance_type: str
    node_instance_id: str
    node_region: str
    node_az: str


def get_kubernetes_cluster_members(job_search_prefix: str) -> List[KubernetesJobMember]:
    kubernetes.config.load_kube_config()
    client = CoreV1Api()

    node_name_to_pod_name = {}
    for pod_info in client.list_namespaced_pod(namespace="default", watch=False).items:
        pod_name: str = pod_info.metadata.name
        node_name: str = pod_info.spec.node_name

        if not pod_name.startswith(job_search_prefix):
            continue

        node_name_to_pod_name[node_name] = pod_name

    cluster_regions = set()
    cluster_azs = set()
    cluster_instance_types = set()

    cluster_members = []
    for node_info in client.list_node().items:
        node_name = node_info.metadata.name
        node_instance_id = json.loads(
            node_info.metadata.annotations["csi.volume.kubernetes.io/nodeid"]
        )["fsx.csi.aws.com"]
        if node_name in node_name_to_pod_name:
            pod_name = node_name_to_pod_name[node_name]
            worker_id = int(re.search(pod_id_pattern, pod_name).group(1))

            node_region = node_info.metadata.labels["topology.kubernetes.io/region"]
            cluster_regions.add(node_region)

            node_az = node_info.metadata.labels["topology.kubernetes.io/zone"]
            cluster_azs.add(node_az)

            node_instance_type = node_info.metadata.labels["node.kubernetes.io/instance-type"]
            cluster_instance_types.add(node_instance_type)

            member_def = KubernetesJobMember(
                pod_name=pod_name,
                worker_id=worker_id,
                node_instance_type=node_instance_type,
                node_instance_id=node_instance_id,
                node_region=node_region,
                node_az=node_az,
            )
            cluster_members.append(member_def)

    if len(cluster_regions) != 1:
        raise RuntimeError(f"found {len(cluster_regions)} regions but only expected 1.")

    if len(cluster_instance_types) != 1:
        print(f"WARNING: found {len(cluster_instance_types)}, generally there should only be 1.")

    if len(cluster_members) != len(node_name_to_pod_name):
        raise RuntimeError(
            f"found {len(node_name_to_pod_name)} pods but "
            f"was only able to match {len(cluster_members)} to instance IDs"
        )

    return cluster_members


def get_free_instances() -> Set[str]:
    kubernetes.config.load_kube_config()
    client = CoreV1Api()

    occupied_nodes = set()
    for pod_info in client.list_namespaced_pod(namespace="default", watch=False).items:
        occupied_nodes.add(pod_info.spec.node_name)

    free_instance_ids = set()
    for node_info in client.list_node().items:
        node_name = node_info.metadata.name
        if node_name not in occupied_nodes:
            node_instance_id = json.loads(
                node_info.metadata.annotations["csi.volume.kubernetes.io/nodeid"]
            )["fsx.csi.aws.com"]
            free_instance_ids.add(node_instance_id)

    return free_instance_ids



"""
https://arxiv.org/pdf/2401.14489
https://developer.nvidia.com/blog/cuda-refresher-cuda-programming-model/
https://docs.nvidia.com/deeplearning/performance/dl-performance-matrix-multiplication/index.html
"""

from typing import Tuple

from .math import ceil_divide, product


def compute_gemm_flops(n_tokens: int, weight_shape: Tuple[int, ...]) -> float:
    """compute the number of FLOPs in a linear layer. Given by 2MNK."""
    return 2 * n_tokens * product(*weight_shape)


def compute_gemm_n_tiles(
    n_tokens: int, weight_shape: Tuple[int, ...], tile_shape: Tuple[int, int, int]
) -> int:
    """get the total number of tiles for a GEMM. GEMMs can be tiled:
    * along output dimension M
    * along output dimension N
    * along reduction dimension K (requires reduction)
    """
    gemm_m = n_tokens
    gemm_k, gemm_n = weight_shape

    # tile_m, tile_n = (256, 128) if gemm_m > gemm_n else (128, 256)
    tile_m, tile_n = 128, 128
    tile_k = gemm_k

    return product(
        ceil_divide(gemm_m, tile_m),
        ceil_divide(gemm_n, tile_n),
        ceil_divide(gemm_k, tile_k),
    )



import dataclasses
import functools
import itertools
from typing import Dict, List, Optional, Set

from botocore.client import BaseClient

import dlcalc.utils.cluster.ec2


@dataclasses.dataclass
class TreeNode:
    node_id: str
    parent: "TreeNode"
    children: Set["TreeNode"]

    def __hash__(self) -> int:
        return hash(self.node_id)


def build_instance_tree(
    ec2_client: BaseClient,
    accepted_node_instance_types: Set[str],
    accepted_node_availability_zones: Set[str],
    accepted_instance_ids: Optional[Set[str]] = None,
) -> Dict[str, TreeNode]:
    node_id_to_node: Dict[str, TreeNode] = {}

    # construct tree of instance topology
    for instance_info in dlcalc.utils.cluster.ec2.iter_instance_info(
        ec2_client=ec2_client,
        accepted_node_instance_types=accepted_node_instance_types,
        accepted_node_availability_zones=accepted_node_availability_zones,
        accepted_instance_ids=accepted_instance_ids,
    ):
        if instance_info.instance_id in node_id_to_node:
            raise AssertionError(
                f"instance {instance_info.instance_id} has been visited twice. "
                f"check for duplicate nodes from EC2 API or cycles in connectivity "
                f"(which are not supported)"
            )

        # order of network nodes: farthest from instance -> closest to instance
        chain = [*instance_info.network_nodes, instance_info.instance_id]

        # update nodes
        for node_id in chain:
            if node_id not in node_id_to_node:
                node_id_to_node[node_id] = TreeNode(node_id, parent=None, children=set())

        # update edges
        for parent_node_idx in range(len(chain) - 1):
            src_node_id = chain[parent_node_idx]
            dst_node_id = chain[parent_node_idx + 1]

            # set parent -> child relationship
            parent_node = node_id_to_node[src_node_id]
            child_node = node_id_to_node[dst_node_id]

            if child_node.parent is None:
                child_node.parent = parent_node
            elif child_node.parent != parent_node:
                raise AssertionError("mismatched parents")  # TODO. better error message
            parent_node.children.add(child_node)

    # some sanity checks
    root_nodes = set()
    nonroot_nodes = set()
    all_nodes = set()
    for node in node_id_to_node.values():
        if node.parent is None:
            root_nodes.add(node)
        else:
            nonroot_nodes.add(node)
        all_nodes.add(node)

    if len(root_nodes) != 1:
        raise RuntimeError(
            f"expected to find 1 root node out of {len(all_nodes)} nodes "
            f"but got {len(root_nodes)}"
        )
    for node in nonroot_nodes:
        assert node.parent in all_nodes

    return node_id_to_node


def dfs_tree_leaves_only(root: TreeNode) -> List[TreeNode]:
    if not root.children:
        return [root]

    list_of_list_of_leaves = []
    for child in root.children:
        list_of_list_of_leaves.append(dfs_tree_leaves_only(child))

    # now we have a list of list of leaves acquired from each subtree.
    # we want this list of lists of leaves to be ordered from largest list -> smallest list
    # this means we use the largest subtrees first - meaning that we will prioritize
    # groups of instances that can communicate with maximum "locality"
    list_of_list_of_leaves.sort(key=len, reverse=True)

    # stick everything together. this gives a prioritized order of instances
    # that should be part of the same ring.
    list_of_leaves = list(itertools.chain.from_iterable(list_of_list_of_leaves))

    return list_of_leaves


@functools.lru_cache(maxsize=None)
def distance_between_nodes(node_a: TreeNode, node_b: TreeNode) -> int:
    if node_a == node_b:
        return 0

    return 2 + distance_between_nodes(node_a.parent, node_b.parent)


def traversal_distance_of_ring(ring_participants: List[TreeNode]) -> int:
    traversed_distance = 0
    for src_node_idx in range(len(ring_participants)):
        dst_node_idx = (src_node_idx + 1) % len(ring_participants)

        src_node = ring_participants[src_node_idx]
        dst_node = ring_participants[dst_node_idx]

        step_distance = distance_between_nodes(src_node, dst_node)

        traversed_distance += step_distance

    return traversed_distance



from enum import Enum


class ActivationCheckpointingType(Enum):
    NONE = 1
    SELECTIVE = 2
    SUPER_SELECTIVE = 3
    FULL = 4

    @staticmethod
    def from_str(str_key: str) -> "ActivationCheckpointingType":
        return {
            "none": ActivationCheckpointingType.NONE,
            "selective": ActivationCheckpointingType.SELECTIVE,
            "super-selective": ActivationCheckpointingType.SUPER_SELECTIVE,
            "full": ActivationCheckpointingType.FULL,
        }[str_key]



"""
determine the best instance placement and rank assignment to ensure DP ring
communication can occur with the fewest possible network hops.
"""

from argparse import ArgumentParser
from typing import List

import boto3

import dlcalc.utils.cluster.ec2
import dlcalc.utils.cluster.kubernetes
import dlcalc.utils.cluster.topology
from dlcalc.utils.cluster.topology import TreeNode
from dlcalc.utils.math import safe_divide


def main() -> None:
    parser = ArgumentParser(__doc__)
    parser.add_argument(
        "-t",
        "--tp-degree",
        type=int,
        required=True,
        help="tensor parallel (TP) degree",
    )
    parser.add_argument(
        "-p",
        "--pp-degree",
        type=int,
        required=True,
        help="pipeline parallel (PP) degree",
    )
    parser.add_argument(
        "-d",
        "--dp-degree",
        type=int,
        required=True,
        help="data parallel (DP) degree",
    )
    parser.add_argument(
        "-n",
        "--accelerators-per-node",
        type=int,
        default=8,
        help="number of accelerators per node",
    )
    parser.add_argument(
        "-i",
        "--instance-type",
        type=str,
        help="required instance type",
    )
    parser.add_argument(
        "-r",
        "--region",
        type=str,
        help="required AWS region",
    )
    parser.add_argument(
        "-a",
        "--az",
        type=str,
        help="required AWS availability zone",
    )
    args = parser.parse_args()
    tp_degree = args.tp_degree
    pp_degree = args.pp_degree
    dp_degree = args.dp_degree
    accelerators_per_node = args.accelerators_per_node

    n_required_nodes = safe_divide(tp_degree * pp_degree * dp_degree, accelerators_per_node)

    free_instance_ids = dlcalc.utils.cluster.kubernetes.get_free_instances()
    print(f"found {len(free_instance_ids)} free instances")
    instance_type = args.instance_type
    region_name = args.region
    az = args.az

    assert tp_degree <= accelerators_per_node

    # we only care to optimize inter-node DP communication
    # (intra-node DP communication can't be optimized with better instance placement),
    # so we'll ignore the parts of the ring that traverse intra-node links.
    # we're making the assumption that all TP groups are within a single node.
    # since we're looking at things at the instance level, there's only one
    # ring per PP rank.
    dp_ranks_per_node = safe_divide(accelerators_per_node, tp_degree)
    nodes_per_dp_ring = safe_divide(dp_degree, dp_ranks_per_node)
    print(
        f"DP={dp_degree} and {dp_ranks_per_node} DP ranks/node, "
        f"therefore there are {nodes_per_dp_ring} nodes per DP ring"
    )
    node_id_to_node = dlcalc.utils.cluster.topology.build_instance_tree(
        ec2_client=boto3.client("ec2", region_name=region_name),
        accepted_node_instance_types={
            instance_type,
        },
        accepted_node_availability_zones={
            az,
        },
        accepted_instance_ids=free_instance_ids,
    )
    n_discovered_instances = len(node_id_to_node)

    if n_discovered_instances < n_required_nodes:
        raise RuntimeError(
            f"training job requires {n_required_nodes} nodes but found {n_discovered_instances}"
        )
    print(f"discovered {len(node_id_to_node)} instances")

    # find root node
    (root_node,) = [node for node in node_id_to_node.values() if node.parent is None]

    # determine optimal topology
    # we want to select an ordered subset of nodes for optimal adjacency.
    # the most adjacent set of nodes for ring collectives minimizes the number
    # of network nodes that need to be crossed.
    # Physical topology is a tree. in each level we can order by the subtree size,
    # then we'll walk it DFS. this will give us an ordered list of instances that would
    # make relatively suitable rings.
    # TODO. this is sort of a heuristic, will not always find the most optimal
    # placement.
    ordered_nodes = dlcalc.utils.cluster.topology.dfs_tree_leaves_only(root_node)

    if len(ordered_nodes) != n_discovered_instances:
        # TODO. need to debug further to see why not all nodes are reachable by the DFS
        # just warning for now.
        print(
            f"[WARN] expected to traverse {n_discovered_instances} instances "
            f"but only encountered {len(ordered_nodes)}"
        )
        # raise RuntimeError(
        #     f"expected to traverse {len(instance_ids)} nodes "
        #     f"but encountered {len(ordered_nodes)}"
        # )

    ring_assignments: List[List[TreeNode]] = [None] * pp_degree
    for pp_rank in range(pp_degree):
        ring_assignments[pp_rank] = ordered_nodes[:nodes_per_dp_ring]
        ordered_nodes = ordered_nodes[nodes_per_dp_ring:]

        print(f"ring for PP={pp_rank}")
        print([n.node_id for n in ring_assignments[pp_rank]])
        traversal_dist = dlcalc.utils.cluster.topology.traversal_distance_of_ring(
            ring_assignments[pp_rank]
        )
        print(
            f"traversal distance: {traversal_dist} -> avg traversal distance {traversal_dist / nodes_per_dp_ring:.2f}"
        )

    # TODO. have this output an MPI rankfile instead
    print("outputting instance list")
    for ring_assignment in ring_assignments:
        for node in ring_assignment:
            print(node.node_id)


if __name__ == "__main__":
    main()



import math


def safe_divide(numerator: int, denominator: int) -> int:
    """return numerator / denominator, or throw an exception if its not divisible."""
    if numerator % denominator != 0:
        valid_denominators = []
        for candidate_denominator in range(1, numerator):
            if numerator % candidate_denominator == 0:
                valid_denominators.append(candidate_denominator)
        raise ValueError(
            f"{numerator} not divisible by {denominator}. "
            f"valid denominators: {valid_denominators}"
        )

    return numerator // denominator


def ceil_divide(numerator: int, denominator: int) -> int:
    """return ceil(numerator / denominator)"""
    return int(math.ceil(numerator / denominator))


def product(*multiplicands: int) -> int:
    """return the product of a sequence of multiplicands."""
    assert len(multiplicands) > 0
    product = 1
    for multiplicand in multiplicands:
        product *= multiplicand

    return product



"""simple utility to convert samples/second to MFU."""

import re
from argparse import ArgumentParser

n_params_pattern = re.compile(r"([\d\.]+)([a-z])")
n_accelerators_pattern = re.compile(r"(\d+)x(\d+)")


def parse_n_params(n_params_str: str) -> int:
    SUFFIX_TO_FACTOR = {
        "m": 1e6,
        "b": 1e9,
        "t": 1e12,
    }
    n_params_str = n_params_str.lower()
    match = re.match(n_params_pattern, n_params_str)

    if not match:
        raise ValueError(f"unable to parse n_params str {n_params_str}")

    prefix = float(match.group(1))

    suffix = match.group(2).lower()
    if suffix not in SUFFIX_TO_FACTOR:
        raise ValueError(
            f"unable to parse n_params str '{n_params_str}' because of unrecognized suffix '{suffix}'"
        )
    factor = SUFFIX_TO_FACTOR[suffix]

    return int(prefix * factor)


def parse_n_accelerators(n_accelerators_str: str) -> int:
    n_accelerators_str = n_accelerators_str.lower()
    if n_accelerators_str.isnumeric():
        return int(n_accelerators_str)
    elif "x" in n_accelerators_str:
        # axb format
        match = re.match(n_accelerators_pattern, n_accelerators_str)
        multiplicand_1 = int(match.group(1))
        multiplicand_2 = int(match.group(2))

        return multiplicand_1 * multiplicand_2
    else:
        raise ValueError(f"unable to parse n_accelerators input '{n_accelerators_str}'")


def main() -> None:
    parser = ArgumentParser(__doc__)
    parser.add_argument(
        "-s",
        "--samples-per-sec",
        type=float,
        required=True,
        help="training throughput in samples/sec",
    )
    parser.add_argument(
        "-l",
        "--seqlen",
        type=int,
        required=True,
        help="sequence length in tokens (i.e. tokens/sample)",
    )
    parser.add_argument(
        "-m",
        "--model-size",
        type=str,
        required=True,
        help="model size (for example 100m, 100b, or 100t)",
    )
    parser.add_argument(
        "-n",
        "--n-accelerators",
        type=str,
        required=True,
        help="number of accelerators used for training. accepts formats like '16' or '2x8'",
    )
    parser.add_argument(
        "-t",
        "--tflops-per-accelerator",
        type=float,
        required=True,
        help="theoretical tflops per accelerator",
    )
    args = parser.parse_args()

    samples_per_sec = args.samples_per_sec
    tokens_per_sample = args.seqlen
    model_size = parse_n_params(args.model_size)
    n_accelerators = parse_n_accelerators(args.n_accelerators)
    flops_per_accelerator = args.tflops_per_accelerator * 1e12

    tokens_per_sec = tokens_per_sample * samples_per_sec
    flops_per_token = 6 * model_size

    # MFU
    achieved_flops = flops_per_token * tokens_per_sec
    theoretical_flops = flops_per_accelerator * n_accelerators
    mfu = achieved_flops / theoretical_flops

    # tokens/day
    tokens_per_day = tokens_per_sec * 60 * 60 * 24

    print(
        f"inputs:\n"
        f"* model size: {model_size * 1e-9:.2f}B parameters\n"
        f"* sequence length: {tokens_per_sample}\n"
        f"* training throughput: {samples_per_sec:.2f} samples/second\n"
        f"* n_accelerators: {n_accelerators}\n"
        f"* FLOPs/accelerator: {flops_per_accelerator * 1e-12:.2f} TFLOPs\n"
    )

    print(f"{tokens_per_day * 1e-9:.2f}B tokens/day")
    print(f"{mfu * 100:.2f}% MFU")


if __name__ == "__main__":
    main()



from typing import Optional, Tuple

from .math import ceil_divide, product, safe_divide


class Size:
    """representation of how much space something takes, which is derived from
    what the datatype is, and how many elements of that datatype there are.
    """

    def __init__(self, numel: int, bits_per_element: int) -> None:
        self._numel = numel
        self._bits_per_element = bits_per_element

    def numel(self) -> int:
        return self._numel

    def bits(self) -> int:
        return self._numel * self._bits_per_element

    def bytes(self) -> int:
        return self._numel * safe_divide(self._bits_per_element, 8)

    def __add__(self, other: "Size") -> "Size":
        if self._bits_per_element != other._bits_per_element:
            raise ValueError(
                f"attempted addition between two sizes with different bits per element "
                f"{self._bits_per_element} vs {other._bits_per_element}"
            )

        return Size(numel=self._numel + other._numel, bits_per_element=self._bits_per_element)

    def __mul__(self, multiplicand: int) -> "Size":
        return Size(self._numel * multiplicand, self._bits_per_element)

    def __rmul__(self, multiplicand: int) -> "Size":
        return Size(self._numel * multiplicand, self._bits_per_element)

    def __repr__(self) -> str:
        return (
            f"numel: {self._numel * 1e-9:.3f} B, "
            f"bytes: {self._numel * safe_divide(self._bits_per_element, 8) / (1024 ** 3):.3f} GiB"
        )


class TensorRepr:
    """Representation of a (possibly distributed) Tensor,
    has an unpartitioned shape, along with partitioning axes and degree.
    """

    def __init__(
        self,
        unpartitioned_shape: Tuple[int, ...],
        partition_dim: Optional[int],
        partition_degree: int,
        bits_per_elt: int,
        enforce_evenly_partitionable: bool = True,
    ) -> None:
        if partition_degree > 1 and partition_dim is None:
            raise RuntimeError(
                f"with partition degree {partition_degree}, must specify partition_dim"
            )

        partition_dim = partition_dim or 0

        if (
            enforce_evenly_partitionable
            and unpartitioned_shape[partition_dim] % partition_degree != 0
        ):
            raise RuntimeError(
                f"dim {partition_dim} of {unpartitioned_shape} not divisible by {partition_degree}"
            )

        self._shape = unpartitioned_shape
        self._numel = product(*unpartitioned_shape)
        self._partition_dim = partition_dim
        self._partition_degree = partition_degree
        self._bits_per_elt = bits_per_elt
        self._bytes_per_elt = safe_divide(bits_per_elt, 8)

    def shape(self, partitioned: bool) -> Tuple[int, ...]:
        if partitioned:
            shape = list(self._shape)
            shape[self._partition_dim] = safe_divide(
                shape[self._partition_dim], self._partition_degree
            )
            return tuple(shape)
        else:
            return self._shape

    def numel(self, partitioned: bool) -> int:
        return self.__get_numel(partitioned=partitioned)

    def size(self, partitioned: bool) -> Size:
        return Size(
            numel=self.__get_numel(partitioned=partitioned), bits_per_element=self._bits_per_elt
        )

    def __get_numel(self, partitioned: bool) -> int:
        if partitioned:
            return ceil_divide(self._numel, self._partition_degree)
        else:
            return self._numel



import pprint
from argparse import ArgumentParser
from dataclasses import dataclass
from typing import Tuple

import torch
from torch import Tensor


@dataclass
class TensorSummary:
    shape: Tuple[int]
    dtype: torch.dtype

    @staticmethod
    def from_tensor(tensor: Tensor) -> "TensorSummary":
        return TensorSummary(tuple(tensor.shape), dtype=tensor.dtype)


def summarize(obj: object) -> object:
    if isinstance(obj, dict):
        return {k: summarize(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [summarize(o) for o in obj]
    elif isinstance(obj, Tensor):
        return TensorSummary.from_tensor(obj)
    elif hasattr(obj, "__dict__"):
        return summarize(obj.__dict__)
    else:
        return obj


def main():
    parser = ArgumentParser(__doc__)
    parser.add_argument("checkpoint_path")
    args = parser.parse_args()
    c = torch.load(args.checkpoint_path)

    summarized = summarize(c)

    pprint.pprint(summarized, width=120)


if __name__ == "__main__":
    main()



"""CLI tool for estimating performance characteristics of 3D parallel training."""

import json
import math
from argparse import ArgumentParser

import yaml

from dlcalc.utils.comms import (
    get_dp_all_gather_bw_term_s,
    get_dp_all_gather_comm_time_s,
    get_dp_all_gather_latency_term_s,
    get_dp_reduce_scatter_bw_term_s,
    get_dp_reduce_scatter_comm_time_s,
    get_dp_reduce_scatter_latency_term_s,
    get_tp_all_gather_comm_time_s,
    get_tp_reduce_scatter_comm_time_s,
)
from dlcalc.utils.compute import compute_gemm_flops
from dlcalc.utils.configurations import ActivationCheckpointingType
from dlcalc.utils.data import Size, TensorRepr
from dlcalc.utils.hardware import MachineSpec
from dlcalc.utils.math import product, safe_divide
from dlcalc.utils.model_3d import ParallelConfig, ThreeDParallelModel
from dlcalc.utils.printing import print_section_header


def main() -> None:
    parser = ArgumentParser(__doc__)
    parser.add_argument("cfg_path", type=str)
    args = parser.parse_args()

    with open(args.cfg_path) as f:
        cfg = yaml.safe_load(f)

    print_section_header("CONFIG")
    print(json.dumps(cfg, indent=2))

    sequence_len = cfg["data"]["seqlen"]
    microbatch_sz = cfg["data"]["microbatch_sz"]
    hidden_sz = cfg["model"]["hidden_sz"]

    model_repr = ThreeDParallelModel(
        parallelism_cfg=ParallelConfig(
            tp=cfg["parallelism"]["tp"],
            pp=cfg["parallelism"]["pp"],
            dp=cfg["parallelism"]["dp"],
            vpp=cfg["parallelism"]["vpp"],
            sp_enabled=cfg["parallelism"]["sp"],
            zero_level=ParallelConfig.ZeroLevel(cfg["parallelism"]["zero_level"]),
        ),
        sequence_len=sequence_len,
        microbatch_sz=microbatch_sz,
        hidden_sz=hidden_sz,
        n_layers=cfg["model"]["n_layers"],
        n_q_heads=cfg["model"]["n_q_heads"],
        n_kv_heads=cfg["model"]["n_kv_heads"],
        head_dim=cfg["model"]["head_dim"],
        inter_sz=cfg["model"]["inter_sz"],
        glu=cfg["model"]["glu"],
        rotary_embed=cfg["model"]["rotary_embeds"],
        dropout=cfg["model"]["dropout"],
        vocab_sz=cfg["model"]["vocab_sz"],
        tie_embeddings=cfg["model"]["tie_embeddings"],
        act_ckpting_type=ActivationCheckpointingType.from_str(
            cfg["performance"]["activation_checkpointing_type"]
        ),
        bucket_size_bytes=int(cfg["parallelism"]["bucket_size_mb"] * 1e6),
    )

    machine_spec = MachineSpec.from_str(cfg["hardware"]["node_type"])
    cluster_size = model_repr.parallelism_cfg.world_size()
    print(machine_spec)
    print("n_devices: ", cluster_size)
    print("n_nodes: ", safe_divide(cluster_size, machine_spec.n_devices))

    ###################################################################################
    # DATA
    ###################################################################################
    print_section_header("DATA")
    gbs = cfg["data"]["gbs"]
    mbs = cfg["data"]["microbatch_sz"]

    bs_per_mp_rank = safe_divide(gbs, model_repr.parallelism_cfg.dp)
    n_microbatches_per_mp_rank = safe_divide(bs_per_mp_rank, mbs)

    print(f"gbs = {gbs}")
    print(f"gbs/pipeline = {bs_per_mp_rank}")
    print(f"n_microbatches/pipeline = {n_microbatches_per_mp_rank}")

    ###################################################################################
    # MEMORY ANALYSIS
    ###################################################################################

    print_section_header("[MEMORY] STATES")
    print(f"total params: {model_repr.get_total_n_params(partitioned=False) * 1e-9:.2f}B")
    print(model_repr.states)

    # activations
    print_section_header("[MEMORY] TRAINING ACTIVATIONS")
    act_size_per_layer_per_inflight_microbatch = (
        model_repr.activation_size_per_microbatch_per_layer()
    )
    print("act/layer/inflight:", act_size_per_layer_per_inflight_microbatch)
    max_inflight_microbatches = model_repr.parallelism_cfg.pp  # 1F1B
    print("max(inflight):", max_inflight_microbatches)
    layers_per_pp_stage = model_repr.layers_per_pp_stage()
    print("layers/pp:", layers_per_pp_stage)
    vpp_multiplier = model_repr.vpp_penalty()
    print(f"VPP memory multiplier = {vpp_multiplier:.2f}")
    act_memory = (
        act_size_per_layer_per_inflight_microbatch
        * min(n_microbatches_per_mp_rank, max_inflight_microbatches)
        * math.ceil(vpp_multiplier * layers_per_pp_stage)
    )
    print(
        f"act/pp_stage = "
        f"per_microbatch_per_layer_per_inflight * "
        f"{max_inflight_microbatches} * "
        f"{math.ceil(vpp_multiplier * layers_per_pp_stage)} = "
        f"{act_memory}"
    )

    print_section_header("[MEMORY] TOTAL")
    print(
        f"total mem (GiB) = {(model_repr.states.total_bytes(partitioned=True) + act_memory.bytes()) / (1024 ** 3):.3f}GiB"
    )

    ###################################################################################
    # PERF ANALYSIS
    ###################################################################################
    print_section_header("GEMMs (note numbers calculated for 100% flops+bandwidth utilization)")
    for proj_name, weight_repr in [
        ("QKV", model_repr.qkv_weight),
        ("ATTN_OUT", model_repr.attn_out_weight),
        ("MLP1", model_repr.mlp_up_weight),
        ("MLP2", model_repr.mlp_down_weight),
    ]:
        weight_repr: TensorRepr  # type: ignore[no-redef]
        flops = compute_gemm_flops(
            n_tokens=model_repr.sequence_len * model_repr.microbatch_sz,
            weight_shape=weight_repr.shape(partitioned=True),
        )
        print(
            f"{proj_name} {weight_repr.shape(partitioned=False)} --tp--> {weight_repr.shape(partitioned=True)}"
        )
        print(
            f"\tCOMPUTE: {flops * 1e-12:.2f} TFLOPs -> "
            f"{flops/machine_spec.device_spec.peak_flops * 1000:.3f} ms"
        )
        bytes_per_element = model_repr.bits_per_parameter // 8
        gemm_input_dim, gemm_output_dim = weight_repr.shape(partitioned=True)
        weight_bytes = bytes_per_element * weight_repr.numel(partitioned=True)
        input_bytes = bytes_per_element * product(
            model_repr.sequence_len, model_repr.microbatch_sz, gemm_input_dim
        )
        output_bytes = bytes_per_element * product(
            model_repr.sequence_len, model_repr.microbatch_sz, gemm_output_dim
        )
        print(
            f"\tLOAD INPUT: {input_bytes * 1e-9:.2f} GB -> "
            f"{input_bytes / (machine_spec.device_spec.mem_bandwidth_bytes_per_sec) * 1000:.3f} ms"
        )
        print(
            f"\tLOAD_WEIGHT: {weight_bytes * 1e-9:.2f} GB -> "
            f"{weight_bytes / (machine_spec.device_spec.mem_bandwidth_bytes_per_sec) * 1000:.3f} ms"
        )
        print(
            f"\tSTORE_OUTPUT: {output_bytes * 1e-9:.2f} GB -> "
            f"{output_bytes / (machine_spec.device_spec.mem_bandwidth_bytes_per_sec) * 1000:.3f} ms"
        )

    print_section_header("TP COMMUNICATION")
    # TODO. assumes SP, analysis pretty similar if not SP though
    activation_size = Size(
        numel=sequence_len * microbatch_sz * hidden_sz,
        bits_per_element=model_repr.bits_per_parameter,
    )
    print(
        f"TP all-gather: {activation_size}: {get_tp_all_gather_comm_time_s(size=activation_size, n_participants=model_repr.parallelism_cfg.tp, machine_spec=machine_spec) * 1000:.3f} ms"
    )
    print(
        f"TP reduce-scatter: {activation_size}: {get_tp_reduce_scatter_comm_time_s(size=activation_size, n_participants=model_repr.parallelism_cfg.tp, machine_spec=machine_spec) * 1000:.3f} ms"
    )

    print_section_header("PP COMMUNICATION")
    activation_send_time_s = (
        activation_size.bytes() / machine_spec.inter_node_connect.unidirectional_bw_bytes_per_sec
    )
    print(f"PP send/recv: {activation_size}: {activation_send_time_s * 1000:.3f} ms")

    print_section_header("PIPELINE BUBBLE")

    vpp = cfg["parallelism"]["vpp"]
    bs_per_mp_rank = safe_divide(gbs, model_repr.parallelism_cfg.dp)
    n_microbatches_per_mp_rank = safe_divide(bs_per_mp_rank, mbs)

    print(f"VPP pipeline bubble multiplier = {(1 / vpp):.2f}")
    print(
        f"pipeline bubble fraction = {(1 / vpp) * (model_repr.parallelism_cfg.pp - 1) / n_microbatches_per_mp_rank:.2f}"
    )

    print_section_header("DP COMMUNICATION")
    if model_repr.parallelism_cfg.zero_level != ParallelConfig.ZeroLevel.PARTITION_OPTIMIZER:
        raise NotImplementedError
    else:
        ###############################################################################
        # compute the backward time for a single microbatch.
        ###############################################################################
        devices_in_pp_stage_flops = (
            model_repr.parallelism_cfg.tp * machine_spec.device_spec.peak_flops
        )

        # divide by single pipeline stage TFLOPs, since its just for single
        # microbatch there's only one active pipeline stage at a time
        single_microbatch_fwd_time = (
            model_repr.get_single_microbatch_fwd_flops() / devices_in_pp_stage_flops
        )
        single_microbatch_bwd_time = (
            model_repr.get_single_microbatch_bwd_flops() / devices_in_pp_stage_flops
        )
        print(
            f"single PP rank: (if 100% FLOPs utilization):\n"
            f"* single microbatch fwd compute time {single_microbatch_fwd_time * 1000:.3f} ms\n"
            f"* single microbatch bwd compute time {single_microbatch_bwd_time * 1000:.3f} ms"
        )
        print()

        ###############################################################################
        # DP comm times
        ###############################################################################
        # grads are reduced in full-precision
        # params are all-gathered in half-precision
        mp_params_size = model_repr.states.params_shard.size(partitioned=True)
        print(f"params per MP degree: {mp_params_size}")
        # TODO. precisions here assume we are doing AMP
        grad_bucket_numel = model_repr.grad_bucket_numel()
        grad_bucket_size = Size(
            numel=grad_bucket_numel,
            bits_per_element=model_repr.bits_per_grad,
        )
        param_bucket_size = Size(
            numel=grad_bucket_numel,
            bits_per_element=model_repr.bits_per_parameter,
        )
        n_buckets = mp_params_size.numel() / grad_bucket_numel
        print(f"reduce_scatter/all_gather n_buckets = ceil({n_buckets})")
        print()
        # full BW should be divided along all MP ranks within a single node, since
        # they are each participating in their own DP collectives. We make the
        # assumption here that TP is the only form of MP we do within node.
        mp_degree_in_node = model_repr.parallelism_cfg.tp
        grad_bucket_reduce_scatter_lat_term_s = get_dp_reduce_scatter_latency_term_s(
            model_repr.parallelism_cfg.dp,
            machine_spec=machine_spec,
        )
        grad_bucket_reduce_scatter_bw_term_s = get_dp_reduce_scatter_bw_term_s(
            grad_bucket_size,
            n_participants=model_repr.parallelism_cfg.dp,
            mp_degree_in_node=mp_degree_in_node,
            machine_spec=machine_spec,
        )
        grad_bucket_reduce_scatter_time_s = get_dp_reduce_scatter_comm_time_s(
            size=grad_bucket_size,
            n_participants=model_repr.parallelism_cfg.dp,
            mp_degree_in_node=mp_degree_in_node,
            machine_spec=machine_spec,
        )
        print(
            f"reduce_scatter(1_grad_bucket):\n"
            f"\tlatency term = {grad_bucket_reduce_scatter_lat_term_s * 1000:.3f} ms\n"
            f"\tbw term = {grad_bucket_reduce_scatter_bw_term_s * 1000:.3f} ms (if 100% BW utilization)\n"
            f"\tTOTAL = {grad_bucket_reduce_scatter_time_s * 1000:.3f} ms\n"
        )
        param_bucket_all_gather_lat_term_s = get_dp_all_gather_latency_term_s(
            model_repr.parallelism_cfg.dp,
            machine_spec=machine_spec,
        )
        param_bucket_all_gather_bw_term_s = get_dp_all_gather_bw_term_s(
            param_bucket_size,
            n_participants=model_repr.parallelism_cfg.dp,
            mp_degree_in_node=mp_degree_in_node,
            machine_spec=machine_spec,
        )
        param_bucket_all_gather_time_s = get_dp_all_gather_comm_time_s(
            param_bucket_size,
            n_participants=model_repr.parallelism_cfg.dp,
            mp_degree_in_node=mp_degree_in_node,
            machine_spec=machine_spec,
        )
        print(
            f"all_gather(1_param_bucket):\n"
            f"\tlatency term = {param_bucket_all_gather_lat_term_s * 1000:.3f} ms\n"
            f"\tbw term = {param_bucket_all_gather_bw_term_s * 1000:.3f} ms (if 100% BW utilization)\n"
            f"\tTOTAL = {param_bucket_all_gather_time_s * 1000:.3f} ms\n"
        )

        print(
            f"reduce_scatter(all_grad_buckets) time = {grad_bucket_reduce_scatter_time_s * n_buckets * 1000:.3f} ms "
            f"(if 100% BW utilization)"
        )
        print(
            f"all_gather(all_param_buckets) time = {param_bucket_all_gather_time_s * n_buckets * 1000:.3f} ms "
            f"(if 100% BW utilization)"
        )

    ##################################################################################
    # Iteration Time
    ##################################################################################


if __name__ == "__main__":
    main()
